package com.yuandian.dataflow.statemachine;

import java.nio.ByteBuffer;
import java.util.concurrent.atomic.AtomicLong;

import com.alipay.remoting.exception.CodecException;
import com.alipay.remoting.serialization.SerializerManager;
import com.alipay.sofa.jraft.Closure;
import com.alipay.sofa.jraft.Iterator;
import com.alipay.sofa.jraft.Status;
import com.alipay.sofa.jraft.conf.Configuration;
import com.alipay.sofa.jraft.core.StateMachineAdapter;
import com.alipay.sofa.jraft.entity.LeaderChangeContext;
import com.alipay.sofa.jraft.error.RaftException;
import com.alipay.sofa.jraft.storage.snapshot.SnapshotReader;
import com.alipay.sofa.jraft.storage.snapshot.SnapshotWriter;
import com.yuandian.dataflow.statemachine.closure.GenericClosure;
import com.yuandian.dataflow.statemachine.operate.Operate;
import com.yuandian.dataflow.statemachine.operate.Operate.OperateType;
import com.yuandian.dataflow.statemachine.state.State;
import com.yuandian.dataflow.statemachine.state.WorkerState;

import lombok.extern.slf4j.Slf4j;

/**
 * Counter state machine.
 *
 * @author boyan (boyan@alibaba-inc.com)
 *
 *         2018-Apr-09 4:52:31 PM
 */
@Slf4j
public class StateMachine extends StateMachineAdapter {

    // private static final Logger LOG =
    // LoggerFactory.getLogger(StateMachine.class);

    /**
     * State value 全局使用的唯一状态
     */
    private State state = new State();
    /**
     * Leader term
     */
    private final AtomicLong leaderTerm = new AtomicLong(-1);

    public boolean isLeader() {
        return this.leaderTerm.get() > 0;
    }

    /**
     * Returns current value. 只有Get 操作状态由协议流程决定 Apply
     */
    public State getState() {
        return state;
    }

    @Override
    @SuppressWarnings("unchecked")
    public void onApply(final Iterator iter) {
        while (iter.hasNext()) {

            Operate op = null;
            GenericClosure<Operate> closure = null;
            if (iter.done() != null) {
                
                // leader可以直接从 回调closure里提取operate
                closure = (GenericClosure<Operate>)iter.done(); // 只支持单一个State. 全状态机只支持一种提交
                op = closure.getValue();

            } else {
                 // 非leader 需要从getData反序列化出来后处理
                final ByteBuffer data = iter.getData();
                try {

                    op = SerializerManager.getSerializer(SerializerManager.Hessian2).deserialize(
                            data.array(),
                            Operate.class.getName());
                } catch (CodecException e) {
                    log.info("{}", e.toString());
                }

            }

            if (op == null) {
                log.error("op 为 {}. 存在错误, 可能版本不一致", op);
                continue;
            }

            switch (op.getType()) {

                case PUT_WORKERSTATE:
                    
                    WorkerState ws = op.getValue();
                    log.debug("PUT {}", ws.peerId);
                    state.getWorkers().put(ws.peerId, ws);
                    if (closure != null) {
                        closure.success(op);
                        closure.run(Status.OK());
                    }
                    break;
                case REMOVE:
                    if (closure != null) {
                        closure.success(op);
                        closure.run(Status.OK());
                    }
                    break;
                default:
                    break;

            }
           

            iter.next();
        }
    }

    @Override
    public void onSnapshotSave(final SnapshotWriter writer, final Closure done) {
        return;
    }

    @Override
    public void onError(final RaftException e) {
        log.debug("Raft error: {}", e, e);
    }

    @Override
    public boolean onSnapshotLoad(final SnapshotReader reader) {
        return true;
    }

    @Override
    public void onLeaderStart(final long term) {
        log.debug("onLeaderStart {}", StateFactory.getServerId());
        this.leaderTerm.set(term);

        // 判断是否Master线程还在跑, 如果存在则中断
        if (MasterFactory.getMasterExecute().isAlive()) {
            MasterFactory.getMasterExecute().interrupt();
        }

        
        StateFactory.readIndexState(new GenericClosure<State>() {
            @Override
            public void run(Status status) {

                var ws = state.getWorkers().get(StateFactory.getServerId());
                if (ws == null) {
                    ws = new WorkerState(StateFactory.getServerId());
                }

                StateFactory.applyOperate(new Operate(OperateType.PUT_WORKERSTATE, ws), new GenericClosure<Operate>() {
                    @Override
                    public void run(Status status) {
                        log.debug("master update workerstate: {}", status);
                    }
                });
            }
        });

        // 当成为master时候 必须启动
        MasterFactory.getMasterExecute().start();

        super.onLeaderStart(term);
    }

    @Override
    public void onLeaderStop(final Status status) {
        log.debug("onLeaderStop {}", StateFactory.getCluster().getServerId());
        this.leaderTerm.set(-1);
        // 判断是否Master线程还在跑, 如果存在则中断
        if (MasterFactory.getMasterExecute().isAlive()) {
            MasterFactory.getMasterExecute().interrupt();
        }

        super.onLeaderStop(status);
    }

    @Override
    public void onShutdown() {
        log.debug("onShutdown");
        super.onShutdown();
    }

    @Override
    public void onStartFollowing(LeaderChangeContext ctx) {
        log.debug("[onStartFollowing] {} {}", ctx, StateFactory.getCluster().getServerId());
        try {

            // 判断是否Master线程还在跑, 如果存在则中断
            if (MasterFactory.getMasterExecute().isAlive()) {
                MasterFactory.getMasterExecute().interrupt();
            }

           
            var ws = new WorkerState(StateFactory.getServerId());
            log.debug("my: {} leader id {}", StateFactory.getServerId(), StateFactory.getLeaderId());

            Operate.CallOperate(new Operate(OperateType.PUT_WORKERSTATE, ws), new GenericClosure<Operate>() {
                @Override
                public void run(Status status) {
                    log.info("onStartFollowing CallOperate {} {}", status, this.getResponse());
                }
            });
 
            return;
        } catch (Exception e) {
            log.info("{}", e.toString());
        }

        super.onStartFollowing(ctx);
    }

    @Override
    public void onConfigurationCommitted(Configuration conf) {
        super.onConfigurationCommitted(conf);
    }

    @Override
    
    public void onStopFollowing(LeaderChangeContext ctx) {
        log.debug("{} {}", ctx, StateFactory.getCluster().getServerId());

        var ws = new WorkerState(StateFactory.getServerId());

        log.debug("my: {} leader id {}", StateFactory.getServerId(), StateFactory.getLeaderId());

        var op = new Operate(OperateType.PUT_WORKERSTATE, ws);
        Operate.CallOperate(op, new GenericClosure<Operate>() {
            @Override
            public void run(Status status) {
                log.info("{} {}", status, this.getResponse());
            }
        });

        super.onStopFollowing(ctx);
    }

}
